//************************************************************************
// FileName:        StarLite.c
//                  IEEE 802.15.4 communications 
//
// One of the following macros should be defined in the Build Options
// #define OSC_INTERNAL  // sets the config bits for internal oscillator
// #define OSC_4MHz  	// sets the config bits 4MHz oscillator with 4 X PLL to get 16MHz 
// #define OSC_16MHz  	// sets the config bits 16MHz 
//
// One of the following macros should be defined in the Build Options
// #define CODE_PROTECT  	// code protection on
// #define NO_CODE_PROTECT 	// code protection off
//
//*************************************************************************

// includes

#include <p18cxxx.h>   

#include "MACAPI.h"
#include <delays.h> 
#include <string.h> 

//******************************************************************************

// function prototypes
void TxChar(char ch);
void TxHex(BYTE toPrint);
BYTE TranslateHexChar( BYTE ch );
void TryStartTx( void );
void ConfigMode( void );
void EraseNetworkData( void );
void FlashLed( void );
void ClearLedFlash( void );
void InitUART( void );

//********************************************************************************/
// semaphores


typedef union _APP_FLAGS
{
    BYTE Val;
    struct _Bits
    {   
	    BYTE BindButtonState : 1;
		BYTE NeedToPoll : 1;
		BYTE IamPolling : 1;
		BYTE AllowJoin : 1;
		BYTE FlashLED : 1;
		BYTE LastPacketWasReleased : 1;
		BYTE ThisPacketWasReleased : 1;
		BYTE ResetAfterFlash : 1;
	} Bits;
}  APP_FLAGS;

APP_FLAGS AppFlags = 0b00000011;  

TICK tTemp;
TICK tLastCharArrival = 0;
TICK FlashStart;
TICK ButtonEvent = 0;
TICK PollEvent = 0;
DWORD_VAL dwTemp;

#define SL_Idle				0
#define SL_Uninitialized	1
#define SL_StartingNetwork	2
#define SL_Starting			4
#define SL_JoiningNetwork	5
#define SL_RejoiningNetwork	6
#define SL_NetworkFailure	7			// possibly can ignore, don't necessarily reset
#define SL_SeriousNetworkFailure	8   // cannot ignore, must reset

BYTE StarLiteState = SL_Uninitialized;

WORD_VAL	MyPANid;
BYTE		MyChannel;
WORD_VAL	MyCoordShortAddr;
WORD_VAL 	MyDestinationAddress;

//********************************************************************************/

// nonvolatile data
#define MAX_CHILD 255

#pragma romdata NVdata=0x0340

	#ifdef SLPC
	  rom BYTE nvNumChild = 0;  // Number of children
	  rom LONG_ADDR nvChildAddr[ MAX_CHILD ];  		// Long addresses of children.  Short address of ChildAddr[yy] is implicitly 0x01yy 
	  rom BYTE	 	nvChildIsSleepy[ MAX_CHILD ];  	// Nonzero if child is sleepy 
	#else
	  rom BYTE nvJoined = 0;  // Nonzero if we have already 
	#endif
	
	// Order of definition of following settings is important for config mode
	rom BYTE nvEraseLock = 'U';
	rom BYTE nvAddressing = 'B';
	rom BYTE nvBaudRate = '9';
	rom BYTE nvPollRate = '3';
	rom BYTE nvPersistTime = '3';
	rom BYTE nvReleaseTime = '3';
	rom BYTE nvOnFail = 'C';
	rom BYTE filler1[0x40]; // ensures we don't erase executable code
	
#pragma romdata

//********************************************************************************/

// serial buffers in and out

#ifdef SLPC
#define BUFFER_SIZE 256
#define FLOW_CONTROL_STOP_AT	225
#define FLOW_CONTROL_CLEAR_AT	200
#else
#define BUFFER_SIZE 128
#define FLOW_CONTROL_STOP_AT	100
#define FLOW_CONTROL_CLEAR_AT	80
#endif

// maximum data in a packet is 64 bytes
#define MAX_TX_PACKET 64

#pragma udata RxBuffer
BYTE RxBuff[BUFFER_SIZE];
#pragma udata TxBuffer
BYTE TxBuff[BUFFER_SIZE];
#pragma udata 
BYTE RxStart = 0;
BYTE RxNext = 0;
BYTE CharsInRx = 0;
BYTE TxStart = 0;
BYTE TxNext = 0;

// temporary stores
BYTE cTemp;
BYTE cTemp2;
BYTE cTemp3;
BYTE cTemp4;
BYTE cTemp5;
BYTE* pTemp;
BYTE* pTemp2;
ROM BYTE* pRTemp;
WORD_VAL wTemp;

//********************************************************************************/

// Pin definitions
#define RTSpinT 		TRISAbits.TRISA1
#define RTSpin 			PORTAbits.RA1
#define CTSpinT 		TRISAbits.TRISA2
#define CTSpin 			PORTAbits.RA2
#define SleepPinT 		TRISBbits.TRISB4
#define SleepPin 		PORTBbits.RB4
#define StatusPinT 		TRISAbits.TRISA4
#define StatusPin 		PORTAbits.RA4
#ifdef DEBUG_MODE
  #define BindPinT 		TRISBbits.TRISB5 
  #define BindPin 		PORTBbits.RB5
#else
  #define BindPinT 		TRISBbits.TRISB7
  #define BindPin 		PORTBbits.RB7
#endif

//--------------------------------------------------------------------------------------------------------

// MAC state machine

ZIGBEE_PRIMITIVE    currentPrimitive;

#define ALLOWED_CHANNELS 	0x07fff800  // allow all channels
#define SUCCESS				0x00
//--------------------------------------------------------------------------------------------------------

// main program

void main(void) 
{
	// set up internal oscillator for 16MHz operation if required
#ifdef OSC_INTERNAL
	OSCCONbits.IRCF1 = 1;  	// assuming <IDCF2:IDCF0>=100 on startup, changes to <IDCF2:IDCF0>=110 = 4MHz
	OSCTUNEbits.PLLEN = 1;  // PLL 4MHz -> 16MHz
	Delay1KTCYx( 100 ); 	// allow 25ms for clock to settle
#endif

    // Initialize MAC layer
    currentPrimitive = NO_PRIMITIVE; 
    MACAPIInit();
    
    switch (nvPersistTime)
    {
	    case '1': dwTemp.Val = ONE_SECOND; break;
	    case '2': dwTemp.Val = ONE_SECOND * ((DWORD) 10 ); break;
	    default:  dwTemp.Val = ONE_SECOND * ((DWORD) 60 ); break;
	    case '4': dwTemp.Val = ONE_SECOND * ((DWORD) 3600 ); break;
	}
	SetPersistTime( dwTemp );
	
	// initialize I/O. 
    ADCON1 = 0x0F; 		 // Make PORTA digital I/O.
    RTSpin = 1;
    RTSpinT = 0;
    CTSpinT = 1;
    SleepPinT = 1;
    BindPinT = 1;
    StatusPin = 1;
    StatusPinT = 0;

	// go into config mode if mac address not valid
	if (!IsMACAddressValid())  ConfigMode();
	    
	// (Re-)Initialize UART
	InitUART();

   // initialize addressing
    MyDestinationAddress.Val = 0xFFFF;
#ifndef SLPC
	if (nvAddressing=='S') MyDestinationAddress.Val = 0x0000;
#endif
	if (nvAddressing=='A') AppFlags.Bits.LastPacketWasReleased = 1;;

	// enable watchdog timer 
    ClrWdt(); 
    ENABLE_WDT();
 
    while (1)
    { 
		// perform mailbox tasks
	    ClrWdt(); 
        MACAPITasks( &currentPrimitive );
        
		// restart output buffer if CTS has cleared
		TryStartTx();
		
		// service status LED
		ClearLedFlash();
		
        switch (currentPrimitive)
        {
			default:
				// ignore
                currentPrimitive = NO_PRIMITIVE;
                break;

#ifndef SLPC
			case MLME_DISASSOCIATE_indication:
				// dissociate infication will only occur if associated with non-starlite nodes.  Politeness would require we EraseNetworkData( );
				// however, while security is not implemented, this would be a security weakness, so we will ignore it instead
				// EraseNetworkData( );
                currentPrimitive = NO_PRIMITIVE;
				break;
#endif

#ifdef SLPC
			case MLME_ASSOCIATE_indication:
				// whatever the result, turn off allow join
				StatusPin = 0;
				macPIB.macAssociationPermit = 0;
				cTemp2 = (params.MLME_ASSOCIATE_indication.CapabilityInformation.CapBits.RxOnWhenIdle? 0x00 : 0x01);

       			// compile associate response
       			currentPrimitive = MLME_ASSOCIATE_response;
                for (cTemp=0; cTemp<sizeof(LONG_ADDR); cTemp++)
                {
                    params.MLME_ASSOCIATE_response.DeviceAddress.v[cTemp] = params.MLME_ASSOCIATE_indication.DeviceAddress.v[cTemp];
                }
                
                // can we permit it to join?
                if (!AppFlags.Bits.AllowJoin)
                {
	                params.MLME_ASSOCIATE_response.status = ASSOCIATION_DENIED;
                    params.MLME_ASSOCIATE_response.AssocShortAddress.Val = 0xFFFF;
	            }
                else if (nvNumChild==MAX_CHILD)
                {
                    params.MLME_ASSOCIATE_response.status = ASSOCIATION_PAN_AT_CAPACITY;
                    params.MLME_ASSOCIATE_response.AssocShortAddress.Val = 0xFFFF;
	            }
				else
				{
                    params.MLME_ASSOCIATE_response.status = ASSOCIATION_SUCCESS;
                    
                    // if node is already in table, return that short address, otherwise, allocate new short address
                    for (cTemp=0; cTemp<nvNumChild; cTemp++)
                    {
						if (!memcmppgm2ram ( (void*) &params.MLME_ASSOCIATE_response.DeviceAddress, (void rom *) &nvChildAddr[cTemp], sizeof(LONG_ADDR))) break;
	                }

					// address of node 0 is 0x0100, node 1 is 0x0101, node 2 is 0x0102 etc
					params.MLME_ASSOCIATE_response.AssocShortAddress.byte.MSB = 0x01;
					params.MLME_ASSOCIATE_response.AssocShortAddress.byte.LSB = cTemp;

					// add to list if new
					if (cTemp==nvNumChild)
					{
						// add new address
						NVMWrite((void rom *) &nvChildAddr[cTemp], (void*) &params.MLME_ASSOCIATE_indication.DeviceAddress, sizeof(LONG_ADDR));
						NVMWrite((void rom *) &nvChildIsSleepy[cTemp], (void*) &cTemp2, cTemp2);
						
						// increment counter
						cTemp++;
						NVMWrite(&nvNumChild, &cTemp, 1);
					}
				}
				
				// star addressing on coordinator - if this is the first device to join, 
				// it is the default destination address
				if (nvAddressing=='S' && MyDestinationAddress.Val == 0xFFFF) 
				{
					MyDestinationAddress = params.MLME_ASSOCIATE_response.AssocShortAddress;
				}

				// finish turning off allow join
				AppFlags.Bits.AllowJoin = 0;
                break;
#endif

#ifndef SLPC
			case MLME_ASSOCIATE_confirm: 
				if (params.MLME_ASSOCIATE_confirm.status == SUCCESS)
				{
					// log as joined
					cTemp = 0x01;
					NVMWrite(&nvJoined, &cTemp, 1);
					
					// device initialization complete, can extinguish LED and set RTS to clear
					RTSpin = 0;
					StatusPin = 0;
					StarLiteState = SL_Idle;
	                currentPrimitive = NO_PRIMITIVE;
				}
				else
				{				
					// failed to join, unset pan id
				 	StarLiteState = SL_SeriousNetworkFailure;
					macPIB.macPANId.Val = 0;
					MLME_SET_macPANId_hw();
					
	                currentPrimitive = NO_PRIMITIVE;
	            }
                break;
#endif

#ifdef SLPC
			case MLME_ORPHAN_indication:
               	// iis it one of mine?
               	for (cTemp=0; cTemp<nvNumChild; cTemp++)
               	{
					if (!memcmppgm2ram ( (void*) &params.MLME_ASSOCIATE_indication.DeviceAddress, (void rom *) &nvChildAddr[cTemp], sizeof(LONG_ADDR))) break;
	            }
	
				if (nvNumChild && cTemp<nvNumChild)
				{
					// mine
					// flash LED to indicate connection
					FlashLed();

					// address of node 0 is 0x0100, node 1 is 0x0101, node 2 is 0x0102 etc
	                params.MLME_ORPHAN_response.ShortAddress.byte.MSB = 0x01;
	                params.MLME_ORPHAN_response.ShortAddress.byte.LSB = cTemp;
	                params.MLME_ORPHAN_response.AssociatedMember = TRUE;
	                // TODO how to set security?  params.MLME_ORPHAN_response.SecurityEnable

					// star addressing on coordinator - if this is the first device to join, 
					// it is the default destination address
					if (nvAddressing=='S' && MyDestinationAddress.Val == 0xFFFF) 
					{
						MyDestinationAddress = params.MLME_ORPHAN_response.ShortAddress;
					}
				}
				else
				{
					// not mine
	                params.MLME_ORPHAN_response.ShortAddress.byte.MSB = 0xFF;
	                params.MLME_ORPHAN_response.ShortAddress.byte.LSB = 0xFF;
	                params.MLME_ORPHAN_response.AssociatedMember = FALSE;
	                // TODO how to set security?  params.MLME_ORPHAN_response.SecurityEnable
				}

                currentPrimitive = MLME_ORPHAN_response;
                break;
#endif

#ifdef SLPC
			case MLME_START_confirm:
				if (params.MLME_START_confirm.status == SUCCESS)
				{
					// device initialization complete, can extinguish LED and set RTS to clear
					RTSpin = 0;
					StatusPin = 0;
					StarLiteState = SL_Idle;
				}			
                currentPrimitive = NO_PRIMITIVE;
                break;
#endif

			case MLME_POLL_confirm:
				AppFlags.Bits.IamPolling = 0;
				if (params.MLME_POLL_confirm.status==NO_ACK) StarLiteState = SL_SeriousNetworkFailure; // no response from parent
                currentPrimitive = NO_PRIMITIVE;
                break;

    		case MLME_BEACON_NOTIFY_indication: 
	      		// If we are scanning for a network to join and the beacon indicates that this device allows joins, note it so
	      		// we can join later.  If multiple networks allow joining, we will end up joining to the last to respond
	      		if (StarLiteState==SL_JoiningNetwork && params.MLME_BEACON_NOTIFY_indication.SuperframeSpec.bits.AssociationPermit)
	      		{
		      		MyPANid = params.MLME_BEACON_NOTIFY_indication.CoordPANId;
		      		MyCoordShortAddr.Val = params.MLME_BEACON_NOTIFY_indication.CoordAddress.ShortAddr.Val;
		      		MyChannel = params.MLME_BEACON_NOTIFY_indication.LogicalChannel;
		      	}
 
				// note how beacon memory buffers MUST be freed with MACDiscardRx( ) - you get beacon notifications during scans 
				MACDiscardRx( );
                currentPrimitive = NO_PRIMITIVE;
                break;

			case MCPS_DATA_indication:
				// Add to TX output buffer
				while (params.MCPS_DATA_indication.msduLength)
				{
					params.MCPS_DATA_indication.msduLength--;
					TxChar( *params.MCPS_DATA_indication.msdu++ );
				}
				// note how data memory buffers always MUST be freed with MACDiscardRx( ) 
				MACDiscardRx( );
                currentPrimitive = NO_PRIMITIVE;
                
                // if star addressing on coordinator, use this address for future transmissions
#ifdef SLPC
                if (nvAddressing=='S' && params.MCPS_DATA_indication.SrcAddrMode == 0x02) 
                	MyDestinationAddress.Val = params.MCPS_DATA_indication.SrcAddr.ShortAddr.Val;
#endif         
       
                // if I am polling for data and data was returned, there may be more, so issue another poll
                // if this packet was a broadcast, then it wan't the reply to the poll; that's still to come
				if (AppFlags.Bits.IamPolling && params.MCPS_DATA_indication.DstAddr.ShortAddr.Val!=0xFFFF)
				{
					AppFlags.Bits.IamPolling = 0;
					AppFlags.Bits.NeedToPoll = 1;
				}
                break;

			case MLME_SCAN_confirm: 
                currentPrimitive = NO_PRIMITIVE;
#ifndef SLPC
				if (params.MLME_SCAN_confirm.ScanType == MAC_SCAN_ORPHAN_SCAN)
				{
					if (!params.MLME_SCAN_confirm.status)
					{
						// successful, nothing to do except indicate status
						RTSpin = 0;
						StatusPin = 0;
						StarLiteState = SL_Idle;
		                currentPrimitive = NO_PRIMITIVE;
					}
					else  
					{				
						// failed to join
	              	  	currentPrimitive = NO_PRIMITIVE;
						StarLiteState = SL_SeriousNetworkFailure;
	              	}
				}
#endif
				if (params.MLME_SCAN_confirm.ScanType == MAC_SCAN_ACTIVE_SCAN)
				{
					if (!params.MLME_SET_confirm.status || params.MLME_SET_confirm.status==MAC_NO_BEACON) 
					{
#ifdef SLPC
#define Channel cTemp
#define iDesc cTemp2
#define NwkCount cTemp3
#define BestNwkCount cTemp4
#define pDesc pTemp
#define pDescC ((PAN_DESCRIPTOR*) pDesc)
	
						// Coordinator - start a network
						
						// loop through descriptor list to select the channel with the minimum number of networks operating
						if (params.MLME_SET_confirm.status!=MAC_NO_BEACON)
						{
							for (Channel = 11; Channel <= 26; Channel++) 
							{
								NwkCount = 0;
								pDesc = (BYTE*) params.MLME_SCAN_confirm.PANDescriptorList;
								for (iDesc = params.MLME_SCAN_confirm.ResultListSize; iDesc != 0; iDesc--)
								{
									if (pDescC->LogicalChannel == Channel) NwkCount++;
									pDesc = (BYTE*) pDescC->next;
								}
								if (Channel==11 || NwkCount < BestNwkCount) 
								{
									MyChannel = Channel;
									BestNwkCount = NwkCount;
								}
							}
							
							// get a unique PAN ID; the following code generates an id which is unique over
							// all frequencies, which is not necessary, but is actually a bit easier to calculate...
							while (1)
							{
								MyPANid.byte.LSB = RANDOM_LSB;
								MyPANid.byte.MSB = RANDOM_MSB & 0x3F;
								pDesc = (BYTE*) params.MLME_SCAN_confirm.PANDescriptorList;
								for (iDesc = params.MLME_SCAN_confirm.ResultListSize; iDesc != 0; iDesc--)
								{
									if (pDescC->CoordPANId.Val == MyPANid.Val) break;
									pDesc = (BYTE*) pDescC->next;
								}
								if (iDesc==0) break;				
							}
						}
						else
						{
							MyChannel = 11;  // Pick the first channel, no special reason
							MyPANid.byte.LSB = RANDOM_LSB;
							MyPANid.byte.MSB = RANDOM_MSB & 0x3F;
						}
	
						// free up memory
						if (params.MLME_SCAN_confirm.ScanType != MAC_SCAN_ORPHAN_SCAN)
						{
							// delete descriptor list
		                    while ((BYTE *)params.MLME_SCAN_confirm.PANDescriptorList)
		                    {
		                        BYTE  * ptr = (BYTE *)params.MLME_SCAN_confirm.PANDescriptorList->next;
		                        SRAMfree( (unsigned char *) params.MLME_SCAN_confirm.PANDescriptorList );
		                        params.MLME_SCAN_confirm.PANDescriptorList = (PAN_DESCRIPTOR *)ptr;
		                    }
						}

						// set the short address to zero 
						macPIB.macShortAddress.Val = 0;
						MLME_SET_macShortAddress_hw();
						
						// to complete network formation
						StarLiteState = SL_Starting;
						currentPrimitive = MLME_START_request;
						params.MLME_START_request.fields.Val = 0x01;  // change to 0x09 when security implemented
						params.MLME_START_request.PANId = MyPANid;
						params.MLME_START_request.LogicalChannel = MyChannel;
						params.MLME_START_request.BeaconOrder = 0x0F;
						params.MLME_START_request.SuperframeOrder = 0x0F;
						
#else

						// end device - if association permitted, request to associate
						if (MyChannel)
						{
	                        // Set the PAN ID to the PAN that we are trying to join; association will complete on receipt of join confirm
							macPIB.macPANId.Val = MyPANid.Val;
							MLME_SET_macPANId_hw();
							
		                    params.MLME_ASSOCIATE_request.SecurityEnable = 0;
		                    params.MLME_ASSOCIATE_request.LogicalChannel = MyChannel;
		                    params.MLME_ASSOCIATE_request.CoordAddrMode = 0x02; // short address
		                    params.MLME_ASSOCIATE_request.CoordAddress.ShortAddr.Val = MyCoordShortAddr.Val;
		                    params.MLME_ASSOCIATE_request.CoordPANId.Val = MyPANid.Val;
  #ifdef SLLS
							params.MLME_ASSOCIATE_request.CapabilityInformation.Val = 0x80; // Sleepy - RFD, no security, Rx off, battery powered, give me a short address
  #else
							params.MLME_ASSOCIATE_request.CapabilityInformation.Val = 0x8E; // Fast - FFD, no security, Rx on, mains powered, give me a short address
  #endif
		                    currentPrimitive = MLME_ASSOCIATE_request;
						}  
						else
						{				
							// failed to join
		              	  	currentPrimitive = NO_PRIMITIVE;
							StarLiteState = SL_SeriousNetworkFailure;
		              	}
#endif
					}
					else
					{				
						// failed to join
	              	  	currentPrimitive = NO_PRIMITIVE;
						StarLiteState = SL_SeriousNetworkFailure;
	              	}
	            }
                break;
                
            case MCPS_DATA_confirm:
            	// flash LED if fail to deliver, then reset
            	if (params.MCPS_DATA_confirm.status)  StarLiteState = SL_NetworkFailure;
                currentPrimitive = NO_PRIMITIVE;
                break;
        }
 
 
		// Pump the stack tasks until idle
		if (currentPrimitive!=NO_PRIMITIVE)  continue;
		
		if (StarLiteState == SL_Uninitialized && BindPin /* don't initialize if button down for config mode */ )
		{
#ifdef SLPC
			// if I am a coordinator and I have not yet started a network, do so now
			StarLiteState = SL_StartingNetwork;
			currentPrimitive = MLME_SCAN_request;
			params.MLME_SCAN_request.ScanType = MAC_SCAN_ACTIVE_SCAN;
			params.MLME_SCAN_request.ScanDuration = 8;
			params.MLME_SCAN_request.ScanChannels.Val = ALLOWED_CHANNELS;
			continue;
#else
			if (nvJoined)
			{
				// if I am an end device and I am not orphan, scan for a old network
				StarLiteState = SL_RejoiningNetwork;
				currentPrimitive = MLME_SCAN_request;
				params.MLME_SCAN_request.ScanType = MAC_SCAN_ORPHAN_SCAN;
				params.MLME_SCAN_request.ScanDuration = 8;
				params.MLME_SCAN_request.ScanChannels.Val = ALLOWED_CHANNELS;
				continue;
			}
			
			// if I am an end device and I am not orphan, scan for network to join
			StarLiteState = SL_JoiningNetwork;
			currentPrimitive = MLME_SCAN_request;
			params.MLME_SCAN_request.ScanType = MAC_SCAN_ACTIVE_SCAN;
			params.MLME_SCAN_request.ScanDuration = 8;
			params.MLME_SCAN_request.ScanChannels.Val = ALLOWED_CHANNELS;
			MyChannel = 0x00;  // will get changed to a channel if association permitted
			continue;
#endif
		}
		    
		if (StarLiteState == SL_NetworkFailure || StarLiteState == SL_SeriousNetworkFailure)
		{
			// sleep ?
			if (nvOnFail=='S' || SleepPin)
			{
				// turn LED off
				StatusPin = 0;

				// sleep	
				MACDisable();
				RBIF = 0;
				RBIE = 1;
				DISABLE_WDT();
				Sleep();
				Nop();
			}
			if (nvOnFail!='C' || StarLiteState == SL_SeriousNetworkFailure) AppFlags.Bits.ResetAfterFlash = 1;  
  			StarLiteState = SL_Idle;
  			FlashLed();
		}
		
		// process button - bear in mind button is active low so BindButtonState is 0 when pressed
		tTemp = TickGet();
		dwTemp.Val = TickGetDiff( tTemp, ButtonEvent );
		if (dwTemp.Val > ONE_SECOND/50 /* 20ms debounce */ && BindPin!=AppFlags.Bits.BindButtonState)
		{
			ButtonEvent = TickGet(); 
			
			// Enter / leave bind state
			AppFlags.Bits.BindButtonState ^= 1;
			if (AppFlags.Bits.BindButtonState && StarLiteState == SL_Idle)  // button up and I have initialized
			{
#ifdef SLPC
				// button must have been released prior to erase, so allow bind
				AppFlags.Bits.AllowJoin ^= 1;
				StatusPin = AppFlags.Bits.AllowJoin;
				macPIB.macAssociationPermit = AppFlags.Bits.AllowJoin;
#endif
			} 
		}
		
		// if held down for longer than five seconds, enter program mode
		tTemp = TickGet();
		dwTemp.Val = TickGetDiff( tTemp, ButtonEvent );
		if ( !AppFlags.Bits.BindButtonState && dwTemp.Val > 5*ONE_SECOND ) 
		{
			// flash LED but don't start config mode until released
			// this is largely for the benefit of operating with DEBUG_MODE set
			while (!BindPin)
			{
				tTemp = TickGet();
				if (tTemp.byte.b1 & 0x30)
					StatusPin = 0;
				else
					StatusPin = 1;
				ClrWdt(); 
			}

			ConfigMode( );
		}

		// Don't start any other processes until StarLite is Idle
		if (StarLiteState != SL_Idle)  continue;
		
		// should we send data?
		tTemp = TickGet();
		dwTemp.Val = TickGetDiff( tTemp, tLastCharArrival );
		cTemp = 0;
		AppFlags.Bits.ThisPacketWasReleased = 1;
		switch (nvReleaseTime)
		{
			case '1':  
				if (dwTemp.Val > ONE_SECOND/((DWORD)1000) && (dwTemp.byte.HIMSB & 0x80) == 0 )  cTemp = CharsInRx;
				break;
			case '2':  
				if (dwTemp.Val > ONE_SECOND/((DWORD)333) && (dwTemp.byte.HIMSB & 0x80) == 0 )  cTemp = CharsInRx;
				break;
			default:  // 10 ms
				if (dwTemp.Val > ONE_SECOND/((DWORD)100) && (dwTemp.byte.HIMSB & 0x80) == 0 )  cTemp = CharsInRx;
				break;
			case '4':  
				if (dwTemp.Val > ONE_SECOND/((DWORD)33) && (dwTemp.byte.HIMSB & 0x80) == 0 )  cTemp = CharsInRx;
				break;
			case '5':  
				if (dwTemp.Val > ONE_SECOND/((DWORD)10) && (dwTemp.byte.HIMSB & 0x80) == 0 )  cTemp = CharsInRx;
				break;
		}
		if (cTemp==0 && CharsInRx>=MAX_TX_PACKET) 
		{
			AppFlags.Bits.ThisPacketWasReleased = 0;
			cTemp = MAX_TX_PACKET;
		}

		// send data
		if (cTemp && IsMACReady() && (nvAddressing!='A' || !AppFlags.Bits.LastPacketWasReleased || cTemp>2) )
		{
			// set address is address mode
			if (AppFlags.Bits.LastPacketWasReleased && nvAddressing=='A') 
			{
				MyDestinationAddress.byte.LSB = RxBuff[RxStart++];
#if (BUFFER_SIZE<256)
				if (RxStart==BUFFER_SIZE) RxStart = 0;
#endif
				MyDestinationAddress.byte.MSB = RxBuff[RxStart++];
#if (BUFFER_SIZE<256)
				if (RxStart==BUFFER_SIZE) RxStart = 0;
#endif

//#define TEST_ADDRESSED_MODE			
#ifdef TEST_ADDRESSED_MODE			
			// typing addresses into hyperterminal is a bit tricky, so this converts characters AA to 0x0000, BB to 0x0100 and CC to 0x0101
			if ( MyDestinationAddress.byte.LSB == 'A' && 	MyDestinationAddress.byte.MSB == 'A' )	MyDestinationAddress.Val = 0x0000;
			if ( MyDestinationAddress.byte.LSB == 'B' && 	MyDestinationAddress.byte.MSB == 'B' )	MyDestinationAddress.Val = 0x0100;
			if ( MyDestinationAddress.byte.LSB == 'C' && 	MyDestinationAddress.byte.MSB == 'C' )	MyDestinationAddress.Val = 0x0101;
#endif			

				cTemp -= 2;
				CharsInRx -= 2;
				AppFlags.Bits.LastPacketWasReleased = AppFlags.Bits.ThisPacketWasReleased;
			}			
			
			MACBlockTx();
			currentPrimitive = MCPS_DATA_request;
		    params.MCPS_DATA_request.SrcAddrMode = 0x02;  // short address
    		params.MCPS_DATA_request.SrcPANId = macPIB.macPANId;
    		params.MCPS_DATA_request.SrcAddr.ShortAddr.Val = macPIB.macShortAddress.Val;
    		params.MCPS_DATA_request.DstAddrMode = 0x02;
    		params.MCPS_DATA_request.DstPANId = macPIB.macPANId;
	    	params.MCPS_DATA_request.msduHandle = 0;
    		params.MCPS_DATA_request.TxOptions.Val = 0x00;
		    params.MCPS_DATA_request.frameType = FRAME_DATA;
    		params.MCPS_DATA_request.DstAddr.ShortAddr.Val = MyDestinationAddress.Val;

		    if (MyDestinationAddress.Val != 0xFFFF)
		    {
		        params.MCPS_DATA_request.TxOptions.bits.acknowledged_transmission = 1;
#ifdef SLPC
		        if (nvChildIsSleepy[ MyDestinationAddress.byte.LSB ])
				{
	                params.MCPS_DATA_request.TxOptions.bits.indirect_transmission = 1;
		        }
#endif
		        // TODO params.MCPS_DATA_request.TxOptions.bits.security_enabled_transmission = ???
		    }
			// Fill transmit buffer
	    	params.MCPS_DATA_request.msduLength = cTemp;
			pTemp = pTxData;
			while (cTemp)
			{
				*pTemp++ = RxBuff[RxStart++];
#if (BUFFER_SIZE<256)
				if (RxStart==BUFFER_SIZE) RxStart = 0;
#endif
				CharsInRx--; 
				cTemp--;
			}

			// release flow control
			if (CharsInRx < FLOW_CONTROL_CLEAR_AT) RTSpin = 0;
			
			continue;
		}


		
#if (defined( SLLS ) )  // only SLLS devices are allowed to sleep

        // If we don't have to execute a primitive, see if we need to request data from
        // our parent, or if we can go to sleep.
        if (currentPrimitive == NO_PRIMITIVE && SleepPin==1  && !AppFlags.Bits.IamPolling)
        {
	        if (StackCanSleep())
			{
				if (TxStart==TxNext)  // check finished outputting data
				{
                  	MACDisable();
				    RTSpin = 1;
					StatusPin = 0;
    	            RBIF = 0;
    	            RBIE = 1;
					DISABLE_WDT();

					Sleep();
              		Nop();

					ClrWdt(); 
					ENABLE_WDT();
					
					MACEnable();
				    RTSpin = 0;
				    AppFlags.Bits.NeedToPoll = 1;
				}
			}
      	}
      	
		// poll if haven't polled for a while
		tTemp = TickGet();
		dwTemp.Val = TickGetDiff( tTemp, PollEvent );
		switch (nvPollRate)
		{
			case '1': if (dwTemp.Val > ONE_SECOND / ((DWORD) 4 ) )  AppFlags.Bits.NeedToPoll = 1;  break;
			case '2': if (dwTemp.Val > ONE_SECOND / ((DWORD) 1 ) )  AppFlags.Bits.NeedToPoll = 1;  break;
			case '3': if (dwTemp.Val > ONE_SECOND )  AppFlags.Bits.NeedToPoll = 1;  break;
			case '4': if (dwTemp.Val > ONE_SECOND * ((DWORD) 2 ) )  AppFlags.Bits.NeedToPoll = 1;  break;
			case '5': if (dwTemp.Val > ONE_SECOND * ((DWORD) 5 ) )  AppFlags.Bits.NeedToPoll = 1;  break;
			case '6': if (dwTemp.Val > ONE_SECOND * ((DWORD) 15 ) )  AppFlags.Bits.NeedToPoll = 1;  break;
			case '7': if (dwTemp.Val > ONE_SECOND * ((DWORD) 60 ) )  AppFlags.Bits.NeedToPoll = 1;  break;
		}
		if ( !AppFlags.Bits.IamPolling && AppFlags.Bits.NeedToPoll )
		{
			PollEvent = tTemp;
			AppFlags.Bits.NeedToPoll = 0;
			AppFlags.Bits.IamPolling = 1;
			currentPrimitive = MLME_POLL_request;
		
           //TODO set this up params.MLME_POLL_request.SecurityEnabled = ???;
           params.MLME_POLL_request.CoordAddrMode          = 0x02;
           params.MLME_POLL_request.CoordPANId             = macPIB.macPANId;
           params.MLME_POLL_request.CoordAddress.ShortAddr = macPIB.macCoordShortAddress;
		}
		
#endif

   }  // end of while (1) loop
}


/*******************************************************************************
Serial tx and rx routines
*******************************************************************************/

void TryStartTx( void )
{
	char ch = CTSpin;
	ch = TXIE;
	if (TxStart!=TxNext && CTSpin==0 && !TXIE)
	{
		// more to transmit
		TXREG = TxBuff[TxStart];
		TxStart++;
#if (BUFFER_SIZE<256)
		if (TxStart==BUFFER_SIZE) TxStart = 0;
#endif
		TXIE = 1;
	}
}

void TxChar(char ch)
{
	cTemp = TxNext;
    TxBuff[TxNext] = ch;
    TxNext++;  
    
#if (BUFFER_SIZE<256)
	if (TxNext==BUFFER_SIZE) TxNext = 0;
#endif
        
    // overflow checking
    if (TxNext==TxStart)
	{
		// Overflow!
		TxNext = cTemp;
		FlashLed();
	}

	TryStartTx();
}


/*******************************************************************************
User Interrupt Handlers

The stack uses some interrupts at low priority for its internal processing.  Once it is done
checking for its interrupts, the stack calls UserInterruptHandler function to allow for any
additional interrupt processing.  Interrupts should not be disabled unless the sleep state 
has been entered.  

PriorityUserInterruptHandler must be for very fast, defined duration operations only

*******************************************************************************/


#pragma interrupt PriorityUserInterruptHandler
void PriorityUserInterruptHandler(void)
{
	static BYTE InChar;
	if (RCIF)
	{
		InChar = RCREG;  
		if ( OERR || FERR ) // overflow or framing error, reset UART
		{
			CREN = 0;
			CREN = 1;
		}
		else
		{
			RxBuff[RxNext] = InChar;
			RxNext++;
			CharsInRx++;
			
#if (BUFFER_SIZE<256)
			if (RxNext==BUFFER_SIZE) RxNext = 0;
#endif

			tLastCharArrival = TickGet();
	
			// flow control		
			if (CharsInRx > FLOW_CONTROL_STOP_AT) RTSpin = 1;
		}
	}
	// clear if interrupt on change
	else if ( RBIE && RBIF )  
	{
        RBIE = 0;
        RBIF = 0;
	}
}


void UserInterruptHandler(void)
{
	// transmission is lower priority than reception
	if (TXIE && TXIF)
	{
		if (TxStart!=TxNext && CTSpin==0)
		{
			// more to transmit
			TXREG = TxBuff[TxStart];
			TxStart++;
#if (BUFFER_SIZE<256)
			if (TxStart==BUFFER_SIZE) TxStart = 0;
#endif
		}
		else
		{
			TXIE = 0;
		}
	}
} 


 
void MACAPIHook( void )
{
}

//------------------------------------------------------

#define N_CONFIG_CHAR   27
#define NUM_SETTINGS	7
#define KEY_LENGTH		4

void ConfigMode( void )
{
#ifdef DEBUG_MODE
  #ifdef SLPC
	BYTE Str[N_CONFIG_CHAR+1] = "#G^9UB9535R0015C83841000000";
  #elif defined SLLF
	BYTE Str[N_CONFIG_CHAR+1] = "#G^9UB9535R0015C83841000006";
  #else
	BYTE Str[N_CONFIG_CHAR+1] = "#G^9UB9535R0015C83841000002";
  #endif
#else
	BYTE Str[N_CONFIG_CHAR+1];
#endif
#define pMACAddrRAM Str	
		
#ifndef DEBUG_MODE
	// 9600 baud, release flow control
	InitUART();
	BAUDCONbits.BRG16 = 0;
	SPBRG = 103;  
	RCIE = 0;
	TXIE = 0;
    RTSpin = 0;
	
	// erase neighbor data if required
	if (nvEraseLock=='U') EraseNetworkData();

	// Send '>' character
	TXREG = '>';
	
	// read in data & flash LED
	cTemp = RCREG;
	pTemp = Str;
	for (cTemp = N_CONFIG_CHAR; cTemp!=0; cTemp--)
	{
		while (1)
		{
			tTemp = TickGet();
			if (tTemp.byte.b1 & 0x30)
				StatusPin = 0;
			else
				StatusPin = 1;
			ClrWdt(); 
			if (RCIF)
			{
				*pTemp++ = RCREG;
				break;
			}
		}
	}
	 
	// Send '<' character
	TXREG = '<';
	while (!TXIF) ClrWdt(); 
#endif	

	// Verify key
	pTemp = Str;
	if ( *pTemp++ != '#' || *pTemp++ != 'G' || *pTemp++ != '^' || *pTemp != '9' ) return;

	// erase neighbor data whatever
	EraseNetworkData();

	// Write first ten bytes to NV memory
	NVMWrite(&nvEraseLock, Str+KEY_LENGTH, NUM_SETTINGS);
	
	// set MAC address if required
	if (!IsMACAddressValid())
	{
		pTemp = pMACAddrRAM;
		for (pTemp2 = &Str[N_CONFIG_CHAR-1]; pTemp2 > &Str[NUM_SETTINGS+KEY_LENGTH]; )
		{
			cTemp = *pTemp2 - '0';  			// converts digits 0-9
			if (cTemp > 0x09) cTemp -=0x07;  	// converts digits A-F
			pTemp2--;
			cTemp2 = *pTemp2 - '0';  			// converts digits 0-9
			if (cTemp2 > 0x09) cTemp2 -=0x07;  	// converts digits A-F
			pTemp2--;
			*pTemp++ = (cTemp2 << 4) + cTemp;
		}
		SetMACAddress(pMACAddrRAM); 
	}

	Reset();
}

void EraseNetworkData( void )
{
	cTemp = 0;
#ifdef SLPC
	NVMWrite(&nvNumChild, &cTemp, 1);
#else
	NVMWrite(&nvJoined, &cTemp, 1);
#endif	
}


void FlashLed( void )
{
	FlashStart = TickGet();
	StatusPin = 1;
	AppFlags.Bits.FlashLED = 1;
	if (AppFlags.Bits.ResetAfterFlash) RTSpin = 1;
}

void ClearLedFlash( void )
{
	tTemp = TickGet();
	if (AppFlags.Bits.FlashLED && tTemp.Val > FlashStart.Val + ONE_SECOND/4)
	{
		StatusPin = 0;
		AppFlags.Bits.FlashLED = 0;
		if (AppFlags.Bits.ResetAfterFlash) Reset();
	}
}

void InitUART( void )
{
	    // Initialize UART 
    switch (nvBaudRate)
    {
	    default: // 9600 baud
		    BAUDCONbits.BRG16 = 0;
		    SPBRG = 103;  
		    break;
	    case '2':  // 19200 baud
		    BAUDCONbits.BRG16 = 1;
		    SPBRG = 207;  
		    break;
	    case '3':  // 38400 baud
		    BAUDCONbits.BRG16 = 1;
		    SPBRG = 103;  
		    break;
	    case '5':  // 57600 baud
		    BAUDCONbits.BRG16 = 1;
		    SPBRG = 68;  
		    break;
	    case '1':  // 115200 baud
		    BAUDCONbits.BRG16 = 1;
		    SPBRG = 34;  
		    break;
	}
    TXSTA = 0x24;
    RCSTA = 0x90; // 0b10010000;
    SPBRGH = 0;
    RCIP = 1;
    RCIE = 1;
    TXIP = 0;
    TXIE = 0;
 	INTCON2bits.RBIP = 1;
    RBIE = 0;
}
